/*
 * gsampler-client-base.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
	 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
	 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n-lib.h>
#include <stdlib.h>
#include <string.h>

#include "gsampler-client.h"
#include "gsampler-client-connection.h"
#include "gsampler-client-device.h"
#include "gsampler-client-channel.h"
#include "gsampler-client-midimap.h"
#include "gsampler-client-database.h"
#include "gsampler-utils.h"
#include "gsampler-option.h"

G_DEFINE_TYPE (GsamplerClient, gsampler_client, G_TYPE_OBJECT);

enum {
	CLIENT_ECHO_EVENT,
	CLIENT_EVENT,
	CLIENT_SERVER_EVENT,
	CLIENT_SERVER_START_EVENT,
	CLIENT_SERVER_ERROR_EVENT,
	CLIENT_DEVICE_COUNT_EVENT,
	CLIENT_DEVICE_INFO_EVENT,
	CLIENT_CHANNEL_COUNT_EVENT,
	CLIENT_CHANNEL_STREAM_EVENT,
	CLIENT_CHANNEL_INFO_EVENT,
	CLIENT_CHANNEL_VOICE_EVENT,
	CLIENT_MAP_COUNT_EVENT,
	CLIENT_MAP_INFO_EVENT,
	CLIENT_MIDI_INSTRUMENT_COUNT_EVENT,
	CLIENT_MIDI_INSTRUMENT_INFO_EVENT,
	CLIENT_TOTAL_STREAM_EVENT,
	CLIENT_TOTAL_VOICE_EVENT,
	CLIENT_DB_PATH_COUNT_EVENT,
	CLIENT_SESSION_UPDATED,
	CLIENT_NB_SIGNALS
};

static guint gsampler_client_signals[CLIENT_NB_SIGNALS] = { 0 };
static GsamplerClient* gsampler_client_instance = NULL;

/**
 * 	Server Routine
 */
static gint 
gsampler_client_server_check () 
{
	gint result = 0;
	gchar* output = NULL;
	GError * error = NULL;

#ifdef G_OS_UNIX
	g_spawn_command_line_sync ("ps -C linuxsampler -o pid=", &output, NULL, NULL, &error);	
	if (error != NULL) 
	{
		g_warning (_("Can't test if linuxsampler run : %s\n"), error->message);
		g_free (output);
		g_error_free (error);
		return 0;
	}
#else
	g_spawn_command_line_sync ("tasklist.exe /nh /fi \"imagename eq linuxsampler.exe\"", 
	                           &output, NULL, NULL, &error);
	if (error != NULL) 
	{
		g_warning (_("Can't test if linuxsampler run : %s\n"), error->message);
		g_free (output);
		g_error_free (error);
		return 0;
	}
	output = string_replace (output, "linuxsampler.exe", "");
#endif

	if (g_strcmp0 (output, "") >> 0) {
		result = atoi (g_strstrip (output));
	}
	g_free (output);
	return result;
}

static gboolean 
gsampler_client_server_callback (GIOChannel* source, GIOCondition condition, gpointer data) 
{
	char buf[50] = {0};
	gsize bytes_read = (gsize) 0;
	GError * error = NULL;
	GsamplerClient* client = GSAMPLER_CLIENT (data);
	g_return_val_if_fail (client != NULL, FALSE);
	g_return_val_if_fail (source != NULL, FALSE);

	//gdk_threads_enter ();
	if (!g_source_is_destroyed (g_main_current_source ()))
    {
		switch (condition) { 
			case G_IO_IN:
			{
				GIOStatus stat = G_IO_STATUS_NORMAL;
				stat = g_io_channel_read_chars (source, buf, 50, &bytes_read, &error);
				if (error != NULL) {
					g_warning (_("error: %s (%s, %d)"), error->message, 
						       g_quark_to_string (error->domain), error->code);
					g_error_free (error);
					gdk_threads_leave ();
					return FALSE;
				}
				if (stat != G_IO_STATUS_NORMAL) {
					g_warning (_("Can't read chars"));
					gdk_threads_leave ();
					return FALSE;
				}
				if (bytes_read != 0) {
					gchar* str = g_utf8_normalize((gchar *)&buf, bytes_read, G_NORMALIZE_DEFAULT);
					g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_EVENT], 0, str);
					if (client->priv->server->started == 0) {
						if (string_contains (str, ").")) {
							client->priv->server->started = 1;
						} else if (string_contains (str, "ds.")) {
							client->priv->server->started = -1;
						}
					}
					g_free (str);
				}
				break;		
			}
			default:
				g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "CRASH", 
					           g_strdup_printf (_("Linuxsampler child watch crash pid : %d"), (gint) client->priv->server->pid));
				gdk_threads_leave ();
				return FALSE;
		}
	}
	//gdk_threads_leave ();
	return TRUE;
}

static gboolean 
gsampler_client_server_start (GsamplerClient* client, const gchar* sample_com, const gchar* arg) 
{
	gint output = 0;
	gchar** argv = NULL;
	gint argv_size = 0;

	GError * error = NULL;
	g_return_val_if_fail (client != NULL, FALSE);
	g_return_val_if_fail (sample_com != NULL, FALSE);
	g_return_val_if_fail (arg != NULL, FALSE);

	if (arg != NULL) 
	{
		sample_com = g_strdup_printf ("%s --instruments-db-location %s", sample_com, arg);
	}
#ifdef G_OS_WIN32
	sample_com = "linuxsampler.exe";
#endif
	g_shell_parse_argv (sample_com, &argv_size, &argv, &error);
	if (error != NULL) 
	{
		g_warning ("file %s: line %d: error: %s", __FILE__, __LINE__, error->message);
		g_strfreev (argv);
		g_error_free (error);
		return FALSE;
	}
	g_spawn_async_with_pipes (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
	                          NULL, NULL, &client->priv->server->pid, NULL,
	                          &output, NULL, &error);
	if (error != NULL) 
	{
		g_warning ("file %s: line %d: error: %s", __FILE__, __LINE__, error->message);
		g_strfreev (argv);
		g_error_free (error);
		return FALSE;
	}
	g_strfreev (argv);
#ifdef G_OS_UNIX
	client->priv->server->server_chan = g_io_channel_unix_new (output);
#else
	client->priv->server->server_chan = g_io_channel_win32_new_fd (output);
#endif
	g_io_channel_set_close_on_unref (client->priv->server->server_chan, TRUE);
	g_io_channel_set_encoding (client->priv->server->server_chan, NULL, &error);
	if (error != NULL) 
	{
		g_warning ("file %s: line %d: error: %s", __FILE__, __LINE__, error->message);
		g_error_free (error);
		return FALSE;
	}
	g_io_channel_set_buffered (client->priv->server->server_chan, FALSE);
#ifdef G_OS_UNIX
	GIOFlags flags;
	flags = g_io_channel_get_flags (client->priv->server->server_chan);
	flags = flags | G_IO_FLAG_NONBLOCK;
	g_io_channel_set_flags (client->priv->server->server_chan, flags, &error);
	if (error != NULL) 
	{
		g_warning ("error: %s", error->message);
		g_strfreev (argv);
		g_error_free (error);
		return FALSE;
	}
#endif
	client->priv->server->watch_id = g_io_add_watch (client->priv->server->server_chan, 
	                                                 G_IO_IN | G_IO_HUP, 
	                                                 gsampler_client_server_callback, client);
	return TRUE;
}

static gboolean 
gsampler_client_server_stop (GsamplerClient* client) 
{
#ifdef G_OS_WIN32
	GError *error = NULL;
	gchar *output = NULL;
#endif
	g_return_val_if_fail (client != NULL, FALSE);

	if (client->priv->server->server_chan != NULL)
	{
		GIOStatus status = g_io_channel_shutdown (client->priv->server->server_chan, FALSE, NULL);
		if (status == G_IO_STATUS_NORMAL) 
		{
			g_io_channel_unref (client->priv->server->server_chan);		
		} else
		{
			g_warning (_("Can't  shutdown the channel status %d"), status);
		}
		g_source_remove (client->priv->server->watch_id);
	}
	if (client->priv->server->pid != 0)
	{
		g_spawn_close_pid (client->priv->server->pid);
#ifdef G_OS_UNIX
		gint kill_status = kill ((pid_t) client->priv->server->pid, SIGTERM);
		if (kill_status != 0) 
		{
			g_warning (_("Can't kill linuxsampler"));
			return FALSE;
		}
#else
		/*g_spawn_command_line_sync (g_strdup_printf("taskkill /PID %d /F", (int) client->priv->server->pid), 
			                       &output, NULL, NULL, &error);*/
		g_spawn_command_line_sync ("taskkill /IM linuxsampler.exe /F", 
			                       &output, NULL, NULL, &error);
		if (error != NULL) 
		{
			g_warning (_("Can't test if linuxsampler run : %s\n"), error->message);
			g_free (output);
			g_error_free (error);
			return 0;
		}
		output = string_replace (output, "linuxsampler.exe", "");
#endif
	}
	return TRUE;
}

static void 
gsampler_client_server_crash (GPid pid, gint status, gpointer data) 
{
	GsamplerClient* client = GSAMPLER_CLIENT (data);
	g_return_if_fail (client != NULL);

	client->priv->server->started = -1;
	if (!gsampler_client_connection_stop (client))
		g_warning (_("Can't kill the client"));
	client->priv->verrou = TRUE;
	if (!g_source_remove (client->priv->server->watch_id)) 
		g_warning (_("Can't disconnect watch listener"));

	g_spawn_close_pid (pid);
	g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "CRASH", 
	               g_strdup_printf (_("Linuxsampler crash pid : %d, status : %d"), (gint) pid, status));
}

static void 
g_cclosure_user_marshal_VOID__INT_INT (GClosure * closure, GValue * return_value, 
                                       guint n_param_values, const GValue * param_values, 
                                       gpointer invocation_hint, gpointer marshal_data) 
{
	typedef void (*GMarshalFunc_VOID__INT_INT) (gpointer data1, gint arg_1, gint arg_2, gpointer data2);
	register GMarshalFunc_VOID__INT_INT callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) 
	{
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else 
	{
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__INT_INT) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_int (param_values + 1), g_value_get_int (param_values + 2), data2);
}

static void 
g_cclosure_user_marshal_VOID__STRING_STRING (GClosure * closure, GValue * return_value, 
                                             guint n_param_values, const GValue * param_values, 
                                             gpointer invocation_hint, gpointer marshal_data) 
{
	typedef void (*GMarshalFunc_VOID__STRING_STRING) (gpointer data1, const char* arg_1, const char* arg_2, gpointer data2);
	register GMarshalFunc_VOID__STRING_STRING callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) 
	{
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else 
	{
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__STRING_STRING) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_string (param_values + 1), g_value_get_string (param_values + 2), data2);
}

static void 
g_cclosure_user_marshal_VOID__BOXED_INT (GClosure * closure, GValue * return_value, 
                                         guint n_param_values, const GValue * param_values, 
                                         gpointer invocation_hint, gpointer marshal_data) 
{
	typedef void (*GMarshalFunc_VOID__BOXED_INT) (gpointer data1, gpointer arg_1, gint arg_2, gpointer data2);
	register GMarshalFunc_VOID__BOXED_INT callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) 
	{
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else 
	{
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__BOXED_INT) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_boxed (param_values + 1), g_value_get_int (param_values + 2), data2);
}

static void g_cclosure_user_marshal_VOID__STRING_DOUBLE_STRING (GClosure * closure, 
                                                                GValue * return_value, 
                                                                guint n_param_values, 
                                                                const GValue * param_values, 
                                                                gpointer invocation_hint, 
                                                                gpointer marshal_data) 
{
	typedef void (*GMarshalFunc_VOID__STRING_DOUBLE_STRING) (gpointer data1, const char* arg_1, gdouble arg_2, const char* arg_3, gpointer data2);
	register GMarshalFunc_VOID__STRING_DOUBLE_STRING callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 4);
	if (G_CCLOSURE_SWAP_DATA (closure)) 
	{
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else 
	{
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__STRING_DOUBLE_STRING) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_string (param_values + 1), g_value_get_double (param_values + 2), g_value_get_string (param_values + 3), data2);
}

static void 
g_cclosure_user_marshal_VOID__INT_INT_INT (GClosure * closure, GValue * return_value, 
                                           guint n_param_values, const GValue * param_values, 
                                           gpointer invocation_hint, gpointer marshal_data) 
{
	typedef void (*GMarshalFunc_VOID__INT_INT_INT) (gpointer data1, gint arg_1, gint arg_2, gint arg_3, gpointer data2);
	register GMarshalFunc_VOID__INT_INT_INT callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 4);
	if (G_CCLOSURE_SWAP_DATA (closure)) 
	{
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else 
	{
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__INT_INT_INT) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_int (param_values + 1), g_value_get_int (param_values + 2), g_value_get_int (param_values + 3), data2);
}

static void 
gsampler_client_init (GsamplerClient * self) 
{
	self->priv = GSAMPLER_CLIENT_GET_PRIVATE (self);

	self->priv->ctx1 = g_new0 (GsamplerConnnection, 1);
	self->priv->ctx2 = g_new0 (GsamplerConnnection, 1);
	self->priv->server = g_new0 (GsamplerServer, 1);
	self->priv->server->started = 0;
	self->priv->verrou = TRUE;
}

static void 
gsampler_client_finalize (GObject* obj) 
{
	GsamplerClient * self;
	self = GSAMPLER_CLIENT (obj);
	g_free (self->priv->ctx1);
	g_free (self->priv->ctx2);
	g_free (self->priv->server);
	g_io_channel_unref (self->priv->chan);
}

static void 
gsampler_client_class_init (GsamplerClientClass * klass) 
{
	gsampler_client_parent_class = g_type_class_peek_parent (klass);
	G_OBJECT_CLASS (klass)->finalize = gsampler_client_finalize;
	g_type_class_add_private (klass, sizeof (GsamplerClientPrivate));

	gsampler_client_signals [CLIENT_ECHO_EVENT] = g_signal_new ("client_echo_event",
	                                                            GSAMPLER_TYPE_CLIENT,
	                                                            G_SIGNAL_RUN_LAST,
	                                                            0, NULL, NULL,
	                                                            g_cclosure_marshal_VOID__STRING,
	                                                            G_TYPE_NONE,
	                                                            1, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_EVENT] = g_signal_new ("client_event", 
	                                                       GSAMPLER_TYPE_CLIENT,
	                                                       G_SIGNAL_RUN_LAST,
	                                                       0, NULL, NULL,
	                                                       g_cclosure_marshal_VOID__STRING,
	                                                       G_TYPE_NONE,
	                                                       1, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_SERVER_EVENT] = g_signal_new ("client_server_event", 
	                                                              GSAMPLER_TYPE_CLIENT, 
	                                                              G_SIGNAL_RUN_LAST, 
	                                                              0, NULL, NULL, 
	                                                              g_cclosure_marshal_VOID__STRING,
	                                                              G_TYPE_NONE, 
	                                                              1, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_SERVER_START_EVENT] = g_signal_new ("client_server_start_event", 
	                                                                    GSAMPLER_TYPE_CLIENT, 
	                                                                    G_SIGNAL_RUN_LAST,
	                                                                    0, NULL, NULL, 
	                                                                    g_cclosure_user_marshal_VOID__STRING_DOUBLE_STRING,
	                                                                    G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_DOUBLE, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_SERVER_ERROR_EVENT] = g_signal_new ("client_serveur_error_event", 
	                                                                    GSAMPLER_TYPE_CLIENT, 
	                                                                    G_SIGNAL_RUN_LAST, 
	                                                                    0, NULL, NULL, 
	                                                                    g_cclosure_user_marshal_VOID__STRING_STRING, 
	                                                                    G_TYPE_NONE, 
	                                                                    2, G_TYPE_STRING, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_DEVICE_COUNT_EVENT] = g_signal_new ("device_count_event", 
	                                                                    GSAMPLER_TYPE_CLIENT, 
	                                                                    G_SIGNAL_RUN_LAST, 
	                                                                    0, NULL, NULL, 
	                                                                    g_cclosure_user_marshal_VOID__BOXED_INT, 
	                                                                    G_TYPE_NONE, 
	                                                                    2, G_TYPE_STRV, G_TYPE_INT);
	gsampler_client_signals [CLIENT_DEVICE_INFO_EVENT] = g_signal_new ("device_info_event", 
	                                                                   GSAMPLER_TYPE_CLIENT, 
	                                                                   G_SIGNAL_RUN_LAST, 
	                                                                   0, NULL, NULL, 
	                                                                   g_cclosure_user_marshal_VOID__INT_INT, 
	                                                                   G_TYPE_NONE, 
	                                                                   2, G_TYPE_INT, G_TYPE_INT);
	gsampler_client_signals [CLIENT_CHANNEL_COUNT_EVENT] = g_signal_new ("channel_count_event", 
	                                                                     GSAMPLER_TYPE_CLIENT, 
	                                                                     G_SIGNAL_RUN_LAST, 
	                                                                     0, NULL, NULL, 
	                                                                     g_cclosure_marshal_VOID__BOXED, 
	                                                                     G_TYPE_NONE, 
	                                                                     1, G_TYPE_STRV);
	gsampler_client_signals [CLIENT_CHANNEL_STREAM_EVENT] = g_signal_new ("channel_stream_event", 
	                                                                      GSAMPLER_TYPE_CLIENT, 
	                                                                      G_SIGNAL_RUN_LAST, 
	                                                                      0, NULL, NULL, 
	                                                                      g_cclosure_user_marshal_VOID__INT_INT, 
	                                                                      G_TYPE_NONE, 
	                                                                      2, G_TYPE_INT, G_TYPE_INT);
	gsampler_client_signals [CLIENT_CHANNEL_INFO_EVENT] = g_signal_new ("channel_info_event", 
	                                                                    GSAMPLER_TYPE_CLIENT, 
	                                                                    G_SIGNAL_RUN_LAST, 
	                                                                    0, NULL, NULL, 
	                                                                    g_cclosure_marshal_VOID__INT, 
	                                                                    G_TYPE_NONE,
	                                                                    1, G_TYPE_INT);
	gsampler_client_signals [CLIENT_CHANNEL_VOICE_EVENT] = g_signal_new ("channel_voice_event", 
	                                                                     GSAMPLER_TYPE_CLIENT, 
	                                                                     G_SIGNAL_RUN_LAST, 
	                                                                     0, NULL, NULL, 
	                                                                     g_cclosure_user_marshal_VOID__INT_INT, 
	                                                                     G_TYPE_NONE, 
	                                                                     2, G_TYPE_INT, G_TYPE_INT);
	gsampler_client_signals [CLIENT_MAP_COUNT_EVENT] = g_signal_new ("map_count_event", 
	                                                                 GSAMPLER_TYPE_CLIENT, 
	                                                                 G_SIGNAL_RUN_LAST, 
	                                                                 0, NULL, NULL, 
	                                                                 g_cclosure_marshal_VOID__BOXED, 
	                                                                 G_TYPE_NONE, 
	                                                                 1, G_TYPE_STRV);
	gsampler_client_signals [CLIENT_MAP_INFO_EVENT] = g_signal_new ("map_info_event", 
	                                                                 GSAMPLER_TYPE_CLIENT, 
	                                                                 G_SIGNAL_RUN_LAST, 
	                                                                 0, NULL, NULL, 
	                                                                 g_cclosure_marshal_VOID__INT, 
	                                                                 G_TYPE_NONE, 
	                                                                 1, G_TYPE_INT);
	gsampler_client_signals [CLIENT_MIDI_INSTRUMENT_COUNT_EVENT] = g_signal_new ("midi_instrument_count_event", 
	                                                                             GSAMPLER_TYPE_CLIENT, 
	                                                                             G_SIGNAL_RUN_LAST, 
	                                                                             0, NULL, NULL, 
	                                                                             g_cclosure_marshal_VOID__BOXED, 
	                                                                             G_TYPE_NONE, 
	                                                                             1, G_TYPE_POINTER);
	gsampler_client_signals [CLIENT_MIDI_INSTRUMENT_INFO_EVENT] = g_signal_new ("midi_instrument_info_event", 
	                                                                             GSAMPLER_TYPE_CLIENT, 
	                                                                             G_SIGNAL_RUN_LAST, 
	                                                                             0, NULL, NULL, 
	                                                                             g_cclosure_user_marshal_VOID__INT_INT_INT, 
	                                                                             G_TYPE_NONE, 
	                                                                             3, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);
	gsampler_client_signals [CLIENT_TOTAL_STREAM_EVENT] = g_signal_new ("total_stream_event", 
	                                                                    GSAMPLER_TYPE_CLIENT, 
	                                                                    G_SIGNAL_RUN_LAST, 
	                                                                    0, NULL, NULL, 
	                                                                    g_cclosure_marshal_VOID__INT, 
	                                                                    G_TYPE_NONE, 
	                                                                    1, G_TYPE_INT);
	gsampler_client_signals [CLIENT_TOTAL_VOICE_EVENT] = g_signal_new ("total_voice_event", 
	                                                                   GSAMPLER_TYPE_CLIENT, 
	                                                                   G_SIGNAL_RUN_LAST, 
	                                                                   0, NULL, NULL, 
	                                                                   g_cclosure_marshal_VOID__INT, 
	                                                                   G_TYPE_NONE, 
	                                                                   1, G_TYPE_INT);
	gsampler_client_signals [CLIENT_DB_PATH_COUNT_EVENT] = g_signal_new ("db_path_count_event", 
	                                                                     GSAMPLER_TYPE_CLIENT, 
	                                                                     G_SIGNAL_RUN_LAST, 
	                                                                     0, NULL, NULL, 
	                                                                     g_cclosure_marshal_VOID__STRING, 
	                                                                     G_TYPE_NONE, 
	                                                                     1, G_TYPE_STRING);
	gsampler_client_signals [CLIENT_SESSION_UPDATED] = g_signal_new ("session_updated", 
	                                                                 GSAMPLER_TYPE_CLIENT, 
	                                                                 G_SIGNAL_RUN_LAST, 
	                                                                 0, NULL, NULL, 
	                                                                 g_cclosure_marshal_VOID__VOID, 
	                                                                 G_TYPE_NONE, 
	                                                                 0);
}

GsamplerClient* 
gsampler_client_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_CLIENT, NULL);
}

GsamplerClient* 
gsampler_client_get_instance (void) 
{
	if (gsampler_client_instance == NULL) 
	{
		gsampler_client_instance = gsampler_client_new ();
	}
	return g_object_ref (gsampler_client_instance);
}

void 
gsampler_client_callback_event (GsamplerClient *self, const gchar* mes, gpointer data) 
{
	//gchar *mes = NULL;
	//GsamplerClient *self = NULL;
	//GsamplerClientJob *event = (GsamplerClientJob *) data;
	//g_return_val_if_fail (event != NULL, FALSE);

	//self = (GsamplerClient *) event->owner;
	//mes = (gchar*) data;
	g_return_if_fail (self != NULL);
	g_return_if_fail (mes != NULL);

	if (string_contains (mes, "NOTIFY:AUDIO_OUTPUT_DEVICE_COUNT")) 
	{
		gchar** dev_list = NULL;
		dev_list = gsampler_client_get_available_audio_output (self);
		g_signal_emit (self, gsampler_client_signals[CLIENT_DEVICE_COUNT_EVENT], 0, dev_list, 0);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_strfreev (dev_list);
	} else if (string_contains (mes, "NOTIFY:MIDI_INPUT_DEVICE_COUNT")) 
	{
		gchar** dev_list = NULL;
		dev_list = gsampler_client_get_available_midi_input (self);
		g_signal_emit (self, gsampler_client_signals[CLIENT_DEVICE_COUNT_EVENT], 0, dev_list, 1);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_strfreev (dev_list);
	} else if (string_contains (mes, "NOTIFY:CHANNEL_COUNT")) 
	{
		gchar** dev_list = NULL;
		dev_list = gsampler_client_get_available_channel (self);
		g_signal_emit (self, gsampler_client_signals[CLIENT_CHANNEL_COUNT_EVENT], 0, dev_list);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_strfreev (dev_list);
	} else if (string_contains (mes, "NOTIFY:AUDIO_OUTPUT_DEVICE_INFO")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:AUDIO_OUTPUT_DEVICE_INFO:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_DEVICE_INFO_EVENT], 0, 0, id);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:MIDI_INPUT_DEVICE_INFO")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:MIDI_INPUT_DEVICE_INFO:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_DEVICE_INFO_EVENT], 0, 1, id);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:CHANNEL_INFO")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:CHANNEL_INFO:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_CHANNEL_INFO_EVENT], 0, id);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:TOTAL_VOICE_COUNT")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:TOTAL_VOICE_COUNT:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_TOTAL_VOICE_EVENT], 0, id);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:TOTAL_STREAM_COUNT")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:TOTAL_STREAM_COUNT:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_TOTAL_STREAM_EVENT], 0, id);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:VOICE_COUNT")) 
	{
		gchar** output = NULL;
		gchar* tmp = NULL;
		gint id1;
		gint id2;
		tmp = string_replace (mes, "NOTIFY:VOICE_COUNT:", "");
		output = g_strsplit (tmp, " ", 0);
		id1 = atoi (g_strstrip (output[0]));
		id2 = atoi (g_strstrip (output[1]));
		g_signal_emit (self, gsampler_client_signals[CLIENT_CHANNEL_VOICE_EVENT], 0, id1 ,id2);
		g_free (tmp);
		g_strfreev (output);
	} else if (string_contains (mes, "NOTIFY:STREAM_COUNT")) 
	{
		gchar* tmp = NULL;
		gchar** output;
		gint id1;
		gint id2;
		tmp = string_replace (mes, "NOTIFY:STREAM_COUNT:", "");
		output = g_strsplit (tmp, " ", 0);
		id1 = atoi (g_strstrip (output[0]));
		id2 = atoi (g_strstrip (output[1]));
		g_signal_emit (self, gsampler_client_signals[CLIENT_CHANNEL_STREAM_EVENT], 0, id1, id2);
		g_free (tmp);
		g_strfreev (output);
	} else if (string_contains (mes, "NOTIFY:MIDI_INSTRUMENT_MAP_COUNT")) 
	{
		gchar** map_list;
		map_list = gsampler_client_get_available_midi_instrument_map (self);
		g_signal_emit (self, gsampler_client_signals[CLIENT_MAP_COUNT_EVENT], 0, map_list);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_strfreev (map_list);
	} else if (string_contains (mes, "NOTIFY:MIDI_INSTRUMENT_COUNT")) 
	{
		gchar* tmp = NULL;
		gchar** output;
		gint id;
		GSList *inst_list = NULL;
		tmp = string_replace (mes, "NOTIFY:MIDI_INSTRUMENT_COUNT:", "");
		output = g_strsplit (tmp, " ", 0);
		g_free (tmp);
		id = atoi (g_strstrip (output[0]));
		inst_list = gsampler_client_get_available_midi_instruments (self, id);
		g_signal_emit (self, gsampler_client_signals[CLIENT_MIDI_INSTRUMENT_COUNT_EVENT], 0, inst_list);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_slist_free_full (inst_list, g_free);
		g_strfreev (output);
	} else if (string_contains (mes, "NOTIFY:MIDI_INSTRUMENT_MAP_INFO")) 
	{
		gchar* output = NULL;
		gint id;
		output = string_replace (mes, "NOTIFY:MIDI_INSTRUMENT_MAP_INFO:", "");
		id = atoi (g_strstrip (output));
		g_signal_emit (self, gsampler_client_signals[CLIENT_MAP_INFO_EVENT], 0, id);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_free (output);
	} else if (string_contains (mes, "NOTIFY:MIDI_INSTRUMENT_INFO")) 
	{
		gchar* tmp = NULL;
		gchar** output;
		tmp = string_replace (mes, "NOTIFY:MIDI_INSTRUMENT_INFO:", "");
		output = g_strsplit (tmp, " ", 0);
		gint id = atoi (g_strstrip (output[0]));
		gint bank = atoi (g_strstrip (output[1]));
		gint prog = atoi (g_strstrip (output[2]));
		g_signal_emit (self, gsampler_client_signals[CLIENT_MIDI_INSTRUMENT_INFO_EVENT], 0, id, bank, prog);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SESSION_UPDATED], 0);
		g_free (tmp);
		g_strfreev (output);
	} else if (string_contains (mes, "NOTIFY:DB_INSTRUMENT_DIRECTORY_COUNT")) 
	{
		gchar* path;
		path = string_replace (mes, "NOTIFY:DB_INSTRUMENT_DIRECTORY_COUNT:", "");
		g_signal_emit (self, gsampler_client_signals[CLIENT_DB_PATH_COUNT_EVENT], 0, g_strstrip (path));
		g_free (path);
	} else if (string_contains (mes, "NOTIFY:MISCELLANEOUS")) 
	{
		gchar* output = NULL;
		output = string_replace (mes, "NOTIFY:MISCELLANEOUS:", "");
		g_signal_emit (self, gsampler_client_signals[CLIENT_ECHO_EVENT], 0, output);
		g_free (output);
	} else 
	{
		g_signal_emit (self, gsampler_client_signals[CLIENT_ECHO_EVENT], 0, mes);
	}
	//g_free (mes);
}

gboolean 
gsampler_client_start (GsamplerClient* client) 
{
	gint pid, wait;
	gchar *com = NULL;
	gchar* arg = NULL;
	gboolean autostart = FALSE;
	GsamplerOption* option;
	g_return_val_if_fail (client != NULL, FALSE);

	option = gsampler_option_get_instance ();
	wait = gsampler_option_get_integer (option, "Serveur", "serveur-wait");
	autostart = gsampler_option_get_boolean (option, "Serveur", "serveur-autostart");
	com = gsampler_option_get_string (option, "Serveur", "serveur-com");
	arg = gsampler_option_get_string (option, "Serveur", "db-path");
	g_object_unref (option);
	
	g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
	               0.2, _("Check if Linuxsampler is running..."));
	pid = gsampler_client_server_check ();
	if (autostart == FALSE && pid == 0) 
	{
		client->priv->server->started = -1;
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
			           0.0, _("Linuxsampler isn't running set autostart and press restart button"));
		g_free (com);
		g_free (arg);
		return FALSE;
	} else if (pid == 0) 
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog",
		               0.2, _("Check Linuxsampler path..."));
		if (!gsampler_utils_file_exist (com))
		{
			g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
			               0.0, g_strdup_printf (_("%s does not exsist, check config/installation."), com));
			g_free (com);
			g_free (arg);
			return FALSE;
		}
		
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog",
		               0.4, _("Start Linuxsampler..."));
		if (gsampler_client_server_start (client, com, arg) != TRUE) 
		{
			client->priv->server->started = -1;
			g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
			               0.0, _("Can't start Linuxsampler..."));
			g_free (com);
			g_free (arg);
			return FALSE;
		}
		client->priv->server->started = 1;
	} else 
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
		               0.6, _("Linuxsampler is standalone..."));
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_EVENT], 0, 
		               _("Linuxsampler is standalone we can't get it's output...\n"));
		client->priv->server->started = 2;
		client->priv->server->pid = (GPid) pid;
	}
	g_free (com);
	g_free (arg);
	
	client->priv->server->watch_id = g_child_watch_add_full (G_PRIORITY_DEFAULT_IDLE,
	                                                         client->priv->server->pid,
	                                                         gsampler_client_server_crash,
	                                                         g_object_ref (client),
	                                                         g_object_unref);
	g_usleep ((gulong) (wait * 1000000));

	g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
	               0.8, _("Connect the client..."));
	if (!gsampler_client_connection_start (client)) 
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
		               0.0, _("Can't connect the client..."));
		return FALSE;
	}
	g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
	               1.0, _("Gsampler is initialized..."));
	return TRUE;
}

gboolean 
gsampler_client_stop (GsamplerClient* client, gboolean force) 
{
	g_return_val_if_fail (client != NULL, FALSE);

	g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
	               0.9, _("Stop the client..."));
	if (!gsampler_client_connection_stop (client)) 
	{
		g_warning (_("Can't disconnect from linuxsampler"));
		return FALSE;
	}
	if (force || (client->priv->server->started != 2 && client->priv->server->started == 1))
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
		               0.5, _("Kill Linuxsampler..."));
		if (!gsampler_client_server_stop (client)) 
		{
			g_warning (_("Can't stop the serveur"));
			return FALSE;
		} else 
		{
			g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "prog", 
			               0.1, _("Linuxsampler is killed..."));
			client->priv->server->started = 0;
		}
	}
	return TRUE;
}

gboolean
gsampler_client_reset_client (GsamplerClient* client)
{
	g_return_val_if_fail (client != NULL, FALSE);

	if (!gsampler_client_connection_stop (client)) 
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error",
		               0.0, _("Can't disconnect from linuxsampler"));
		return FALSE;
	}
	g_usleep (500000);
	if (!gsampler_client_connection_start (client)) 
	{
		g_signal_emit (client, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
		               0.0, _("Can't connect the client..."));
		return FALSE;
	}

	
	return TRUE;
}

gboolean
gsampler_client_restart (GsamplerClient* client)
{
	g_return_val_if_fail (client != NULL, FALSE);

	if (!gsampler_client_stop (client, TRUE)) 
	{
		return FALSE;
	}
	g_usleep (500000);
	if (!gsampler_client_start (client)) 
	{
		return FALSE;
	}
	return TRUE;
}

/**
 * 	Global Routine
 */

void 
gsampler_client_client_timeout (GsamplerClient* self) 
{
	g_return_if_fail (self != NULL);

	g_usleep ((guint) self->priv->timeout);
}

static gpointer
gsampler_client_send_commande_thread (gpointer data)
{
	GString* ret = NULL;
	gchar* com = NULL;
	gchar* retour = NULL;
	gchar *result = NULL;
	GError *error = NULL;
	GsamplerClientJob *job = (GsamplerClientJob *) data;
	g_return_val_if_fail (job != NULL, NULL);

	GsamplerClient *self = (GsamplerClient *) job->owner;
	if (self == NULL) return NULL;
	
	while (self->priv->verrou) 
	{
		g_usleep (self->priv->timeout);
	}

	self->priv->verrou = TRUE;
	ret = g_string_new ("");
	com = g_strconcat (job->command, "\r\n", NULL);
	g_socket_send (self->priv->ctx2->sock, com, strlen (com), NULL, &error);
	if (error == NULL) 
	{
		retour = g_malloc0 (1024);
		g_socket_receive(self->priv->ctx2->sock, retour, 1024, NULL, &error);
		if (error == NULL && retour != NULL) 
		{
			g_string_append (ret, g_strstrip (string_replace (retour, "\r\n", "")));
			while (!string_contains (retour, "\r\n"))
			{
				g_socket_receive(self->priv->ctx2->sock, retour, 1024, NULL, &error);
				if (error != NULL || retour == NULL) break;
				g_string_append (ret, g_strstrip (string_replace (retour, "\r\n", "")));
			}
			g_free (retour);
		}
	}
	if (error != NULL) 
	{
		g_warning ("file %s: line %d: Can't send commande to the serveur : %s\n",
		           __FILE__, __LINE__, error->message);
		g_error_free (error);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "CRASH",  error->message);
	}
	self->priv->verrou = FALSE;
	if (string_contains (ret->str, "ERR:")) 
	{
		g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "COM", ret->str);
				
	} else
	{
		result = g_strdup (ret->str);
	}
	g_object_unref (self);
	g_free (job->command);
	g_free (job);
	g_free (com);
	g_string_free (ret, TRUE);
	return result;
}

gchar* 
gsampler_client_send_commande (GsamplerClient* self, const gchar* mes) 
{
	GThread *thread;
	GsamplerClientJob *job;
	GError *error = NULL;
	gchar *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (mes != NULL, NULL);

	if (self->priv->server->started == 0)
	{
		g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
		               0.0, _("Gsampler isn't connected to linuxsampler press restart first..."));
		return NULL;
	}
	job = g_new0 (GsamplerClientJob, 1);
	job->owner = g_object_ref_sink (self);
	job->command = g_strdup (mes);
	thread = g_thread_create (gsampler_client_send_commande_thread, job, TRUE, &error);
	if (error != NULL)
	{
		g_warning ("file %s: line %d: Can't create thread : %s\n",
		           __FILE__, __LINE__, error->message);
		g_error_free (error);
	}
	result = g_thread_join (thread);
	return result;
}

gpointer
gsampler_client_client_get_info_thread (gpointer data) 
{
	gchar* com = NULL;
	gchar* retour = NULL;
	gchar **value = NULL;
	GSList *ret = NULL;
	GError * error = NULL;
	GsamplerClientJob *job = (GsamplerClientJob *) data;
	g_return_val_if_fail (job != NULL, NULL);

	GsamplerClient *self = (GsamplerClient *) job->owner;
	if (self == NULL) return NULL;

	while (self->priv->verrou) 
	{
		g_usleep ((guint) self->priv->timeout);
	}
	self->priv->verrou = TRUE;
	com = g_strconcat (job->command, "\r\n", NULL);
	g_socket_send (self->priv->ctx2->sock, com, strlen (com), NULL, &error);
	if (error == NULL) 
	{
		retour = g_malloc0 (1024);
		gboolean pass = TRUE;
		while (pass)
		{
			int length, i;
			g_socket_receive(self->priv->ctx2->sock, retour, 1024, NULL, &error);
			if (error != NULL || retour == NULL) break;
			value = g_strsplit (retour, "\r\n", 0);
			length = g_strv_length (value);
			if (length == 0) break;
			if (string_contains (retour, "ERR:")) 
			{
				g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "COM", retour);
				g_slist_free_full (ret, g_free);
				ret = NULL;
				break;
			}
			for (i = 0; i < length; i++)
			{
				gchar *temp = value[i];
				if (temp != NULL && strlen (temp) != 0)
				{
					if (g_strcmp0 (temp, ".") == 0) 
					{
						pass = FALSE;
						break;
					}
					ret = g_slist_prepend (ret, g_strdup (value[i]));
				}
			}
			g_strfreev (value);
		}
		g_free (retour);
	} else
	{
		g_warning ("file %s: line %d: Can't send commande to the serveur : %s\n",
		           __FILE__, __LINE__, error->message);
		g_error_free (error);
		g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_ERROR_EVENT], 0, "CRASH",  error->message);
	}
	self->priv->verrou = FALSE;
	g_free (com);
	g_object_unref (job->owner);
	g_free (job->command);
	g_free (job);
	return ret;
}

GSList*
gsampler_client_client_get_info (GsamplerClient* self, const gchar* mes)
{
	GThread *thread;
	GsamplerClientJob *job;
	GError *error = NULL;
	GSList *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (mes != NULL, NULL);

	if (self->priv->server->started == 0)
	{
		g_signal_emit (self, gsampler_client_signals[CLIENT_SERVER_START_EVENT], 0, "error", 
		               0.0, _("Gsampler isn't connected to linuxsampler press restart first..."));
		return NULL;
	}
	job = g_new0 (GsamplerClientJob, 1);
	job->owner = g_object_ref_sink (self);
	job->command = g_strdup (mes);
	thread = g_thread_create (gsampler_client_client_get_info_thread, job, TRUE, &error);
	if (error != NULL)
	{
		g_warning ("file %s: line %d: Can't create thread : %s\n",
		           __FILE__, __LINE__, error->message);
		g_error_free (error);
		g_free (job->command);
		g_object_unref (job->owner);
		g_free (job);
		return NULL;
	}
	result = g_thread_join (thread);
	return result;
}

gboolean 
gsampler_client_send_bool (GsamplerClient* self, const gchar* com) 
{
	gboolean result = FALSE;
	gchar* ret = NULL;
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (com != NULL, FALSE);

	ret = gsampler_client_send_commande (self, com);
	if (ret != NULL && string_contains (ret, "OK")) 
	{
		result = TRUE;
	}
	g_free (ret);
	return result;
}

gint
gsampler_client_create (GsamplerClient* self, const gchar* com) 
{
	gint result = -1;
	gchar* ret;
	g_return_val_if_fail (self != NULL, -1);							   
	g_return_val_if_fail (com != NULL, -1);

	ret = gsampler_client_send_commande (self, com);
	if (ret != NULL && string_contains (ret, "OK")) 
	{
		result = atoi (string_slice (ret, 3, 4));
	}	
	g_free (ret);
	return result;
}

gchar*
gsampler_client_get_char (GsamplerClient* self, const gchar* com) 
{
	gchar* result = NULL;
	gchar* ret;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (com != NULL, NULL);

	ret = gsampler_client_send_commande (self, com);
	if (ret != NULL && strlen (ret) != 0) 
	{
		result = g_strdup (ret);
	}
	g_free (ret);
	return result;
}

gint
gsampler_client_get_int (GsamplerClient* self, const gchar* com) 
{
	gint result = -1;
	gchar* ret;
	g_return_val_if_fail (self != NULL, -1);
	g_return_val_if_fail (com != NULL, -1);

	ret = gsampler_client_send_commande (self, com);
	if (ret != NULL && strlen (ret) != 0) 
	{
		result = atoi(ret);
	}
	g_free (ret);
	return result;
}

gchar** 
gsampler_client_get_list (GsamplerClient* self, const gchar* com) 
{
	gchar** result = NULL;
	gchar* ret;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (com != NULL, NULL);

	ret = gsampler_client_send_commande (self, com);
	if (ret != NULL && strlen (ret) != 0) 
	{
		result = g_strsplit (ret, ",", 0);
	}
	g_free (ret);
	return result;
}

gchar* gsampler_client_get_master_volume (GsamplerClient* self) 
{
	return gsampler_client_get_char (self, "GET VOLUME");
}

gboolean 
gsampler_client_set_master_volume (GsamplerClient* self, gdouble vol) 
{
	return gsampler_client_send_bool (self, g_strdup_printf ("SET VOLUME %s", double_to_string ((vol/100))));
}

gboolean gsampler_client_client_reset (GsamplerClient* self) {
	
	return gsampler_client_send_bool (self, "RESET");
}