/*
 * gsampler-device.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
/*#include <math.h>*/
#include <glib/gi18n-lib.h>

#include "gsampler-device.h"
#include "gsampler-utils.h"
#include "gsampler-window.h"
#include "gsampler-option.h"

G_DEFINE_TYPE (GsamplerDevice, gsampler_device, GTK_TYPE_FRAME);

enum {
	GSAMPLER_DEVICE_DUMMY_PROPERTY,
	GSAMPLER_DEVICE_TYPE,
	GSAMPLER_DEVICE_LAST
};

static GParamSpec *dev_pops[GSAMPLER_DEVICE_LAST];

void gsampler_device_set_device_type (GsamplerDevice* self, gint type);

static GsamplerDeviceInfo* 
gsampler_device_get_device_info (GsamplerDevice* self, gint id, gint type) 
{
	GsamplerClient* client = NULL;
	GsamplerDeviceInfo *new_info = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	
	client = gsampler_client_get_instance ();
	new_info = gsampler_client_get_device_info (client, id, type);
	g_object_unref (client);
	if (new_info->id == (-1))
	{
		gsampler_device_info_free (new_info);
		return NULL;
	}
	return new_info;
}

static void 
gsampler_device_on_param_changed (GtkAdjustment* sender, gpointer data) 
{
	GsamplerClient* client;
	gchar* value = NULL;
	gdouble val;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (sender != NULL);
	
	client = gsampler_client_get_instance ();
	val = gtk_adjustment_get_value (sender);
	if (self->priv->info->type == 0) 
	{
		value = double_to_string (val);			
		if (!gsampler_client_set_audio_output_parameter (client, self->priv->info->id, "CHANNELS", value))
		{
			gsampler_main_window_send_message ("warning", _("Can't set audio channel"));
		}
	} else if (self->priv->info->type == 1) 
	{
		value = double_to_string (val);
		if (!gsampler_client_set_midi_input_parameter (client, self->priv->info->id, "PORTS", value)) 
		{
			gsampler_main_window_send_message ("warning", _("Can't set midi port"));
		}
	} else 
	{
		gsampler_main_window_send_message ("warning", g_strdup_printf (_("Invalide type : %d"), self->priv->info->type));
	}
	g_free (value);
	g_object_unref (client);
}

static void 
gsampler_device_on_device_delete (GtkButton* sender, gpointer data) {
	GsamplerClient* client;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (sender != NULL);
	
	client = gsampler_client_get_instance ();
	switch (self->priv->info->type) 
	{
		case 0:
			if (!gsampler_client_delete_audio_device (client, self->priv->info->id)) 
			{
				gsampler_main_window_send_message ("warning",
				                                   g_strdup_printf (_("Can't supress audio device %d"), 
				                                                    self->priv->info->id));
			}
			break;
		case 1:
			if (!gsampler_client_delete_midi_device (client, self->priv->info->id)) 
			{
				gsampler_main_window_send_message ("warning",
				                                   g_strdup_printf (_("Can't supress midi device %d"), 
				                                                    self->priv->info->id));
			}
			break;
		default:
		break;
	}
	g_object_unref (client);
}

static void 
gsampler_device_on_pref_toggled (GtkToggleButton* sender, gpointer data) 
{
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (sender != NULL);

	if (gtk_toggle_button_get_active (sender)) 
		gtk_widget_show ((GtkWidget*) self->priv->pref_box);
	else
		gtk_widget_hide ((GtkWidget*) self->priv->pref_box);
}

static void 
gsampler_device_on_active_toggled (GtkToggleButton* sender, gpointer data)
{
	gboolean val;
	gchar* state = NULL;
	GsamplerClient* client;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (sender != NULL);
	
	val = gtk_toggle_button_get_active (sender);
	client = gsampler_client_get_instance ();
	switch (self->priv->info->type) 
	{
		case 0:
			state = bool_to_string (val);
			if (!gsampler_client_set_audio_output_parameter (client, self->priv->info->id,
			                                                 "ACTIVE", state)) 
			{
				gsampler_main_window_send_message ("warning", _("Can't set active parameter"));
			}
			break;
		case 1:
			state = bool_to_string (val);
			if (!gsampler_client_set_midi_input_parameter (client, self->priv->info->id,
			                                               "ACTIVE", state)) 
			{
				gsampler_main_window_send_message ("warning", _("Can't set active parameter"));
			}
			break;
		default:
		break;
	}
	g_free (state);
	g_object_unref (client);
}

static void 
gsampler_device_on_param_combo_changed (GtkComboBox* sender, gpointer data) 
{
	GtkTreeIter iter = {0};
	GsamplerClient* client = NULL;
	gchar* tmp_param = NULL;
	gint param;
	GtkListStore* liste = NULL;
	GSList *binds = NULL;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (sender != NULL);
	
	client = gsampler_client_get_instance ();
	tmp_param = gsampler_text_combo_get_active (sender);
	if (tmp_param == NULL)
	{
		g_object_unref (client);
		return;
	}
	param = atoi (tmp_param);
	liste = (GtkListStore*) gtk_tree_view_get_model (self->priv->binding_tree);
	gtk_list_store_clear (liste);

	binds = gsampler_client_get_avail_device_parameter_bindings (client, self->priv->info->type, 
	                                                             self->priv->info->id, param);
	for (; binds; binds = binds->next) 
	{
		GsamplerDeviceBinding* binding;
		binding = (GsamplerDeviceBinding*) binds->data;
		gtk_list_store_append (liste, &iter);
		gtk_list_store_set (liste, &iter, 0, binding->active, 1, g_strdup (binding->name), -1);
		gsampler_device_binding_free (binding);
	}
	
	g_free (tmp_param);
	g_object_unref (client);
}


static void 
gsampler_device_on_param_toggled (GtkCellRendererToggle* sender, const gchar* path, gpointer data) 
{
	GtkTreeIter iter = {0};
	GtkTreeIter combo_iter = {0};
	GsamplerClient* client = NULL;
	gchar* param = NULL;
	gint param_id = 0;
	GtkListStore* liste;
	GtkTreePath* tree_path;
	const gchar* tmp_id = NULL;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (path != NULL);
	
	liste = (GtkListStore*) gtk_tree_view_get_model (self->priv->binding_tree);
	tree_path = gtk_tree_path_new_from_string (path);
	gtk_tree_model_get_iter ((GtkTreeModel*) liste, &iter, tree_path);
	gtk_tree_model_get ((GtkTreeModel*) liste, &iter, 1, &param, -1, -1);
	
	GtkTreeModel *combo_model = gtk_combo_box_get_model (self->priv->param_combo);
	if (!gtk_combo_box_get_active_iter (self->priv->param_combo, &combo_iter))
	{
		g_free (param);
		return;
	}
	gtk_tree_model_get (combo_model, &combo_iter, 0 , &tmp_id, -1);
	if (tmp_id == NULL)
	{
		g_free (param);
		return;
	}
	param_id = atoi (tmp_id);
	client = gsampler_client_get_instance ();
	if (self->priv->info->type == 0) 
	{
		if (!gsampler_client_set_audio_output_channel_parameter (client, self->priv->info->id, 
		                                                         param_id, "JACK_BINDINGS", param)) 
		{
			gsampler_main_window_send_message ("warning", _("Can't update audio channel parameter..."));
			gtk_tree_path_free (tree_path);
			g_object_unref (liste);
			g_free (param);
			g_object_unref (client);
			return;
		}
	} else if (self->priv->info->type == 1) 
	{
		gchar* tmp = NULL;
		if (g_strcmp0 (self->priv->info->driver, "JACK") == 0)
		{
			tmp = g_strdup ("JACK_BINDINGS");
		} else 
		{
			tmp = g_strdup ("ALSA_SEQ_BINDINGS");
		}
		if (!gsampler_client_set_midi_input_port_parameter (client, self->priv->info->id, 
		                                                    param_id, tmp, param))
		{
			gsampler_main_window_send_message ("warning", _("Can't update midi port parameter..."));
			g_free (tmp);
			gtk_tree_path_free (tree_path);
			g_object_unref (liste);
			g_free (param);
			g_object_unref (client);
			return;
		}
		g_free (tmp);
	} else 
	{
		gsampler_main_window_send_message ("warning", 
		                                   g_strdup_printf ("Invalide type : %d", self->priv->info->type));
	}
	gsampler_device_on_param_combo_changed ((GtkComboBox*) self->priv->param_combo, self);
	gtk_tree_path_free (tree_path);
	g_free (param);
	g_object_unref (client);
}

#if GTK_CHECK_VERSION (2, 90, 0)
static gboolean
gsampler_device_on_draw_event (GtkWidget* sender, cairo_t* context, gpointer data)
{
#else	
static gboolean
gsampler_device_on_expose_event (GtkWidget* sender, GdkEventExpose* event, gpointer data) 
{
	cairo_t* context = gdk_cairo_create (gtk_widget_get_window(sender));
#endif
	GdkColor red = {0};
	GdkColor green = {0};
	GtkAllocation alloc = {0};
	gdouble xc;
	cairo_pattern_t* pattern;
	PangoFontDescription* font_description;
	PangoLayout* layout;
	gint height = 0;
	gint size = 0;
	gchar* aud = NULL;
	GsamplerDevice* self = GSAMPLER_DEVICE (data);
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (context != NULL, FALSE);
	
	gdk_color_parse ("red", &red);
	gdk_color_parse ("green", &green);
	gtk_widget_get_allocation ((GtkWidget*) sender, &alloc);
	xc = (gdouble) ((0 + alloc.width) / 2);
	cairo_set_source_rgba (context, 0, 0, 0, 1);
	cairo_rectangle (context, 0, 0, alloc.width, alloc.height);
	cairo_fill (context);
	pattern = cairo_pattern_create_linear (xc, 0, xc, alloc.height);
	cairo_pattern_add_color_stop_rgba (pattern, 0, 1, 1, 1, 0);
	cairo_pattern_add_color_stop_rgba (pattern, 0.4, 1, 1, 1, 0.2);
	cairo_pattern_add_color_stop_rgba (pattern, 0.6, 1, 1, 1, 0.2);
	cairo_pattern_add_color_stop_rgba (pattern, 1, 1, 1, 1, 0);
	cairo_set_source (context, pattern);
	cairo_rectangle (context, 0, 0, alloc.width, alloc.height);
	cairo_fill (context);

	font_description = pango_font_description_from_string (self->priv->font);
	layout = pango_cairo_create_layout (context);
	size = abs ((int) ((alloc.height-8)/2));
	height = (int) (pango_font_description_get_size (font_description) / PANGO_SCALE);
	if (height > size) {
		pango_font_description_set_size (font_description, size*PANGO_SCALE);
		height = size;
	}
	pango_layout_set_font_description (layout, font_description);
	
	if (self->priv->info->type == 0) 
	{
		aud = g_strdup (_("Audio "));
	} else if (self->priv->info->type == 1) 
	{
		aud = g_strdup (_("Midi "));
	} else 
	{
		aud = g_strdup (_("Invalide type "));
	}
	
	cairo_set_source_rgb (context, red.red, red.green, red.blue);
	cairo_move_to (context, 10, 1);
	pango_layout_set_text (layout, g_strdup_printf ("%s %i", aud, self->priv->info->id), -1);
	
	pango_cairo_update_layout (context, layout);
	pango_cairo_show_layout (context, layout);
	cairo_set_source_rgb (context, green.red, green.green, green.blue);
	pango_layout_set_text (layout, self->priv->info->driver, -1);
	cairo_move_to (context, (alloc.width - 40), (height+3));
	pango_cairo_update_layout (context, layout);
	pango_cairo_show_layout (context, layout);
	if (self->priv->info->name != NULL) 
	{
		cairo_set_source_rgb (context, 1, 1, 1);
		pango_layout_set_text (layout, self->priv->info->name, -1);
		cairo_move_to (context, 10, (height+3));
		pango_cairo_update_layout (context, layout);
		pango_cairo_show_layout (context, layout);
	}
	if (self->priv->info->type == 0) 
	{
		cairo_set_source_rgb (context, green.red, green.green, green.blue);
		pango_layout_set_text (layout, g_strdup_printf ("%.0f Hz", self->priv->info->samplerate), -1);
		cairo_move_to (context, (gdouble) (alloc.width - 65), 1);
		pango_cairo_update_layout (context, layout);
		pango_cairo_show_layout (context, layout);
	}
	g_free (aud);
	g_object_unref (layout);
	pango_font_description_free (font_description);
	cairo_pattern_destroy (pattern);
	return FALSE;
}

static void 
gsampler_device_get_property (GObject *object, guint property_id,
                              GValue *value, GParamSpec *pspec) {
	GsamplerDevice * self;
	self = GSAMPLER_DEVICE (object);
	g_return_if_fail (self != NULL);
	
	switch (property_id) {
		case GSAMPLER_DEVICE_TYPE:
			g_value_set_int (value, gsampler_device_get_device_type (self));
		break;
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void 
gsampler_device_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec) {
	GsamplerDevice * self;
	self = GSAMPLER_DEVICE (object);
	g_return_if_fail (self != NULL);
	
	switch (property_id) {
		case GSAMPLER_DEVICE_TYPE:
			gsampler_device_set_device_type (self, g_value_get_int (value));
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static GObject *
gsampler_device_constructor (GType type, guint n_construct_properties, 
                             GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GsamplerDevice * self;
	GtkBox* main_box = NULL;
	GtkBox* vbox1 = NULL;
	GtkBox* vbox2 = NULL;
	GtkBox* hbox1 = NULL;
	GtkBox* hbox2 = NULL;
	GtkLabel* param_label = NULL;
	GtkLabel* param_label2 = NULL;
	GtkSpinButton* param_button = NULL;
	GtkToggleButton* pref_button = NULL;
	GtkButton* close_button = NULL;
	GtkImage* pref_img = NULL;
	GtkImage* close_img = NULL;
	GtkSeparator* sep = NULL;
	GtkListStore* list;
	GtkCellRendererToggle* toggle;
	GsamplerOption* option;
	
	obj = G_OBJECT_CLASS (gsampler_device_parent_class)->constructor (type, n_construct_properties, construct_properties);
	self = GSAMPLER_DEVICE (obj);
	gtk_frame_set_shadow_type ((GtkFrame*) self, GTK_SHADOW_ETCHED_OUT);
	gtk_container_set_border_width ((GtkContainer*) self, (guint) 5);

#if GTK_CHECK_VERSION (2, 90, 0)
	gchar* css = "GtkFrame { border-radius: 5; border-style: solid; border-width: 2; border-color: rgba(0, 0, 0, 0.5);}";
	GtkCssProvider *prov = gtk_css_provider_new ();
	if (gtk_css_provider_load_from_data (prov, css, -1, NULL)) {
		GtkStyleContext *style =  gtk_widget_get_style_context((GtkWidget *) self);
		gtk_style_context_add_provider (style, (GtkStyleProvider *) prov, GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
	}
	g_object_unref (prov);
#endif
							   
	self->priv->top_box = (GtkBox*) gtk_vbox_new (FALSE, 0);
	main_box = (GtkBox*) gtk_hbox_new (FALSE, 0);
	vbox1 = (GtkBox*) gtk_vbox_new (FALSE, 5);
	hbox1 = (GtkBox*) gtk_hbox_new (FALSE, 0);
	if (self->priv->info->type == 0) {
		param_label = (GtkLabel*) gtk_label_new (NULL);
		gtk_label_set_markup (param_label, _("<b>Channels :</b>"));
		param_label2 = (GtkLabel*) gtk_label_new (NULL);
		gtk_label_set_markup (param_label2, _("<b>Channel :</b>"));
		self->priv->param_adj = (GtkAdjustment *) gtk_adjustment_new (1, 1, 750, 1, 10, 0);
	} else if (self->priv->info->type == 1) {
		param_label = (GtkLabel*) gtk_label_new (NULL);
		gtk_label_set_markup (param_label, _("<b>Midi ports :</b>"));
		param_label2 = (GtkLabel*) gtk_label_new (NULL);
		gtk_label_set_markup (param_label2, _("<b>Midi port :</b>"));
		self->priv->param_adj = (GtkAdjustment *) gtk_adjustment_new (1, 1, 253, 1, 10, 0);
	}
	
	param_button = (GtkSpinButton*) gtk_spin_button_new (self->priv->param_adj, 0, 0);
	self->priv->draw_area = (GtkDrawingArea*) gtk_drawing_area_new ();
	vbox2 = (GtkBox*) gtk_vbox_new (FALSE, 5);
	pref_img = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_CONNECT, GTK_ICON_SIZE_MENU);
	pref_button = (GtkToggleButton*) gtk_toggle_button_new ();
	gtk_button_set_image ((GtkButton*) pref_button, (GtkWidget*) pref_img);
	self->priv->active_button = (GtkToggleButton*) gtk_toggle_button_new_with_label (_("Active"));
	close_img = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_DELETE, GTK_ICON_SIZE_MENU);
	close_button = (GtkButton*) gtk_button_new ();
	gtk_button_set_image (close_button, (GtkWidget*) close_img);
	sep = (GtkSeparator*) gtk_hseparator_new ();
	self->priv->pref_box = (GtkBox*) gtk_vbox_new (FALSE, 0);
	hbox2 = (GtkBox*) gtk_hbox_new (FALSE, 0);

	self->priv->param_combo = gsampler_text_combo_create ();
	
	list = gtk_list_store_new (2, G_TYPE_BOOLEAN, G_TYPE_STRING);
	self->priv->binding_tree = (GtkTreeView*) gtk_tree_view_new_with_model ((GtkTreeModel*) list);
	toggle = (GtkCellRendererToggle*) gtk_cell_renderer_toggle_new ();
	gtk_tree_view_insert_column_with_attributes (self->priv->binding_tree, -1, _("Active"),
	                                             (GtkCellRenderer*) toggle,
	                                             "active", 0, NULL, NULL);
	gtk_tree_view_insert_column_with_attributes (self->priv->binding_tree, -1, _("Bindings"),
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 1, NULL, NULL);
	gtk_box_pack_start (hbox1, (GtkWidget*) self->priv->active_button, FALSE, TRUE, (guint) 0);
	gtk_box_pack_start (hbox1, (GtkWidget*) param_label, TRUE, TRUE, (guint) 0);
	gtk_box_pack_start (hbox1, (GtkWidget*) param_button, FALSE, TRUE, (guint) 0);
	gtk_box_pack_start (vbox1, (GtkWidget*) self->priv->draw_area, TRUE, TRUE, (guint) 0);
	gtk_box_pack_start (vbox1, (GtkWidget*) hbox1, FALSE, TRUE, 0);
	gtk_box_pack_start (vbox2, (GtkWidget*) close_button, FALSE, TRUE, 1);
	gtk_box_pack_start (vbox2, (GtkWidget*) pref_button, FALSE, TRUE,  1);
	gtk_box_pack_start (main_box, (GtkWidget*) vbox1, TRUE, TRUE, 0);
	gtk_box_pack_start (main_box, (GtkWidget*) vbox2, FALSE, TRUE, 3);
	gtk_box_pack_start (hbox2, (GtkWidget*) param_label2, TRUE, TRUE, 0);
	gtk_box_pack_start (hbox2, (GtkWidget*) self->priv->param_combo, TRUE, TRUE, 0);
	gtk_box_pack_start (self->priv->pref_box, (GtkWidget*) sep, FALSE, FALSE, 5);
	gtk_box_pack_start (self->priv->pref_box, (GtkWidget*) hbox2, FALSE, FALSE, 0);
	gtk_box_pack_start (self->priv->pref_box, (GtkWidget*) self->priv->binding_tree, FALSE, FALSE, 0);
	gtk_box_pack_start (self->priv->top_box, (GtkWidget*) main_box, FALSE, TRUE, 0);
	gtk_box_pack_start (self->priv->top_box, (GtkWidget*) self->priv->pref_box, FALSE, FALSE, 0);
	gtk_container_set_border_width ((GtkContainer*) self->priv->top_box, 5);
	gtk_container_add ((GtkContainer*) self, (GtkWidget*) self->priv->top_box);
	g_signal_connect_object (toggle, "toggled", 
	                         (GCallback) gsampler_device_on_param_toggled, self, 0);
	g_signal_connect_object (close_button, "clicked", 
	                         (GCallback) gsampler_device_on_device_delete, self, 0);
	g_signal_connect_object (pref_button, "toggled", 
	                         (GCallback) gsampler_device_on_pref_toggled, self, 0);
	option = gsampler_option_get_instance ();
	self->priv->font = gsampler_option_get_string (option, "General", "font");
	g_object_unref (option);
#if (GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION < 90) || GTK_MAJOR_VERSION < 2
	g_signal_connect_object ((GtkWidget*) self->priv->draw_area, "expose_event", 
	                         (GCallback) gsampler_device_on_expose_event, self, 0);
#else
	g_signal_connect_object ((GtkWidget*) self->priv->draw_area, "draw", 
	                         (GCallback) gsampler_device_on_draw_event, self, 0);
#endif
	gtk_widget_show_all ((GtkWidget*) self);
	gtk_widget_hide ((GtkWidget*) self->priv->pref_box);
	return obj;
}

static void 
gsampler_device_init (GsamplerDevice * self) 
{
	self->priv = GSAMPLER_DEVICE_GET_PRIVATE (self);
}

static void 
gsampler_device_finalize (GObject* obj) 
{
	GsamplerDevice * self;
	self = GSAMPLER_DEVICE (obj);
	g_object_unref (self->priv->top_box);
	g_object_unref (self->priv->draw_area);
	g_object_unref (self->priv->param_adj);
	g_object_unref (self->priv->active_button);
	g_object_unref (self->priv->pref_box);
	g_object_unref (self->priv->param_combo);
	g_object_unref (self->priv->binding_tree);
	g_free (self->priv->font);
	gsampler_device_info_destroy (self->priv->info);
	G_OBJECT_CLASS (gsampler_device_parent_class)->finalize (obj);
}

static void 
gsampler_device_class_init (GsamplerDeviceClass * klass) 
{
	gsampler_device_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerDevicePrivate));
	G_OBJECT_CLASS (klass)->get_property = gsampler_device_get_property;
	G_OBJECT_CLASS (klass)->set_property = gsampler_device_set_property;
	G_OBJECT_CLASS (klass)->constructor = gsampler_device_constructor;
	G_OBJECT_CLASS (klass)->finalize = gsampler_device_finalize;

	dev_pops[GSAMPLER_DEVICE_TYPE] = g_param_spec_int ("type", "Type", "Gsampler device type",
	                                                   0, 1, 0,
	                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE );
	g_object_class_install_property (G_OBJECT_CLASS (klass),
	                                 GSAMPLER_DEVICE_TYPE, 
	                                 dev_pops[GSAMPLER_DEVICE_TYPE]);
}

GsamplerDevice* 
gsampler_device_new (gint type) 
{
	return g_object_new (GSAMPLER_TYPE_DEVICE, "type", type, NULL);
}

gboolean 
gsampler_device_active (GsamplerDevice* self, gint id) 
{
	g_return_val_if_fail (self != NULL, FALSE);

	gsampler_device_info_free (self->priv->info);
	self->priv->info = gsampler_device_get_device_info (self, id, self->priv->info->type);
	if (self->priv->info == NULL && self->priv->info->id == -1) 
	{
		return FALSE;
	} 
	gtk_adjustment_set_value (self->priv->param_adj, (gdouble) self->priv->info->param);
	g_signal_connect_object (self->priv->param_adj, "value-changed",
	                         (GCallback) gsampler_device_on_param_changed, self, 0);
	if (self->priv->info->active == TRUE) 
	{
		gtk_toggle_button_set_active (self->priv->active_button, TRUE);
	}
	g_signal_connect_object (self->priv->active_button, "toggled", 
	                         (GCallback) gsampler_device_on_active_toggled, self, 0);
	gint i;
	gsampler_text_combo_remove_all (self->priv->param_combo);
	for (i = 0; i < self->priv->info->param; i++) 
	{
		gsampler_text_combo_append (self->priv->param_combo, g_strdup_printf ("%i", i));
	}
	g_signal_connect_object (self->priv->param_combo, "changed", 
	                         (GCallback) gsampler_device_on_param_combo_changed, self, 0);
	gtk_combo_box_set_active (self->priv->param_combo, 0);
	return TRUE;
}

gint 
gsampler_device_get_device_id (GsamplerDevice* self) 
{
	g_return_val_if_fail (self != NULL, -1);
	
	return self->priv->info->id;
}

gint 
gsampler_device_get_device_type (GsamplerDevice* self) 
{
	g_return_val_if_fail (self != NULL, -1);
	
	return self->priv->info->type;
}

void 
gsampler_device_set_device_type (GsamplerDevice* self, gint type) 
{
	g_return_if_fail (self != NULL);

	if (self->priv->info == NULL) 
	{
		self->priv->info = g_new0 (GsamplerDeviceInfo, 1);
	}
	self->priv->info->type = type;
}

GsamplerDeviceInfo *
gsampler_device_get_info (GsamplerDevice* self) 
{
	g_return_val_if_fail (self != NULL, NULL);
	
	return gsampler_device_info_dup (self->priv->info);
}

void 
gsampler_device_update_info (GsamplerDevice* self) 
{
	GsamplerDeviceInfo* new_info = NULL;
	g_return_if_fail (self != NULL);
	
	new_info = gsampler_device_get_device_info (self, self->priv->info->id, self->priv->info->type);
	if (new_info != NULL)
	{
		gint active, i;
		gsampler_device_info_destroy (self->priv->info);
		gsampler_device_info_copy (new_info, self->priv->info);
		gtk_widget_queue_draw ((GtkWidget*) self->priv->draw_area);
		active = gtk_combo_box_get_active (self->priv->param_combo);
		gsampler_text_combo_remove_all (self->priv->param_combo);
		for (i = 0; i < self->priv->info->param; i++) 
		{
			gsampler_text_combo_append (self->priv->param_combo, g_strdup_printf ("%i", i));
		}
		gtk_combo_box_set_active (self->priv->param_combo, active);
	}
	gsampler_device_info_free (new_info);
}